/*******************************************************************************
 * Copyright (C) 2009 SourceForge.JP Hiroyuki Inaba and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Hiroyuki Inaba - initial API and implementation
 *******************************************************************************/
package jp.sourceforge.mylyn.internal.web.tasks;

import java.io.IOException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringEscapeUtils;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.mylyn.internal.web.tasks.Messages;
import org.eclipse.mylyn.internal.web.tasks.NamedPattern;
import org.eclipse.mylyn.internal.web.tasks.TasksWebPlugin;
import org.eclipse.mylyn.internal.web.tasks.Util;
import org.eclipse.mylyn.tasks.core.IRepositoryQuery;
import org.eclipse.mylyn.tasks.core.TaskRepository;
import org.eclipse.mylyn.tasks.core.data.TaskAttributeMapper;
import org.eclipse.mylyn.tasks.core.data.TaskData;
import org.eclipse.mylyn.tasks.core.data.TaskDataCollector;
import org.eclipse.mylyn.tasks.core.data.TaskMapper;
import org.eclipse.mylyn.tasks.core.sync.ISynchronizationSession;

public class WebRepositoryConnector extends
		org.eclipse.mylyn.internal.web.tasks.WebRepositoryConnector {

	public static final String REPOSITORY_TYPE = "jp.web"; //$NON-NLS-1$

	private static final String PARAM_GROUP_ID = PARAM_PREFIX + "group_id"; //$NON-NLS-1$

	private static final String PARAM_PRIORITY_1 = PARAM_PREFIX + "Priority_1"; //$NON-NLS-1$
	private static final String PARAM_PRIORITY_2 = PARAM_PREFIX + "Priority_2"; //$NON-NLS-1$
	private static final String PARAM_PRIORITY_3 = PARAM_PREFIX + "Priority_3"; //$NON-NLS-1$
	private static final String PARAM_PRIORITY_4 = PARAM_PREFIX + "Priority_4"; //$NON-NLS-1$
	private static final String PARAM_PRIORITY_5 = PARAM_PREFIX + "Priority_5"; //$NON-NLS-1$

	private static final String PARAM_COMPLETED_STATUSES = PARAM_PREFIX
			+ "Completed_Statuses"; //$NON-NLS-1$

	private static final String PARAM_DATE_FORMAT = PARAM_PREFIX
			+ "Date_Format"; //$NON-NLS-1$

	private String server_url = ""; //$NON-NLS-1$
	private String group_id = ""; //$NON-NLS-1$
	private String priorities[] = null;
	private String statuses = null;
	private String date_format = null;

	public String getConnectorKind() {
		return REPOSITORY_TYPE;
	}

	public String getLabel() {
		return Messages_jp.WebRepositoryConnector_SourceForge_JP_Tasks_Ticket_;
	}

	public void updateTaskMapper(TaskMapper mapper, TaskRepository repository,
			NamedPattern pattern, Matcher matcher) {

		String server = repository.getRepositoryUrl();
		String group = repository.getProperty(PARAM_GROUP_ID);
		if (group == null)
			group = ""; //$NON-NLS-1$

		if (!(server_url.equals(server) && group_id.equals(group))) {
			server_url = server;
			group_id = group;

			priorities = new String[5];
			priorities[0] = evaluateParams(repository
					.getProperty(PARAM_PRIORITY_1), repository);
			priorities[1] = evaluateParams(repository
					.getProperty(PARAM_PRIORITY_2), repository);
			priorities[2] = evaluateParams(repository
					.getProperty(PARAM_PRIORITY_3), repository);
			priorities[3] = evaluateParams(repository
					.getProperty(PARAM_PRIORITY_4), repository);
			priorities[4] = evaluateParams(repository
					.getProperty(PARAM_PRIORITY_5), repository);
			statuses = evaluateParams(repository
					.getProperty(PARAM_COMPLETED_STATUSES), repository);
			date_format = evaluateParams(repository
					.getProperty(PARAM_DATE_FORMAT), repository);
		}

		String value = pattern.group("Status", matcher); //$NON-NLS-1$
		if (value != null && value.length() > 0) {
			if (statuses != null) {
				if (statuses.contains(value.toLowerCase())) {
					mapper.setCompletionDate(DEFAULT_DATE);
				}
			}
		}

		value = unescapeGroup(pattern, "Open", matcher, repository); //$NON-NLS-1$
		if (value != null && value.length() > 0) {
			try {
				Date date = parseDate(value);
				mapper.setCreationDate(date);

				if (mapper.getCompletionDate() != null) {
					mapper.setCompletionDate(date);
				}
			} catch (Exception e) {
			}
		}

		value = unescapeGroup(pattern, "Update", matcher, repository); //$NON-NLS-1$
		if (value != null && value.length() > 0) {
			try {
				Date date = parseDate(value);
				mapper.setModificationDate(date);

				if (mapper.getCompletionDate() != null) {
					mapper.setCompletionDate(date);
				}
			} catch (Exception e) {
			}
		}

		mapper.setOwner(""); //$NON-NLS-1$
		value = unescapeGroup(pattern, "Owner", matcher, repository); //$NON-NLS-1$
		if (value != null && value.length() > 0) {
			// <a href="/users/xxxxx/">xxxxx</a>
			if (value.toLowerCase().startsWith("<a href=")) { //$NON-NLS-1$
				try {
					int p1 = value.indexOf('>');
					int p2 = value.indexOf('<', p1);
					value = value.substring(p1 + 1, p2).trim();
					mapper.setOwner(value);
				} catch (Exception e) {
				}
			}
		}

		value = unescapeGroup(pattern, "Reporter", matcher, repository); //$NON-NLS-1$
		if (value != null && value.length() > 0) {
			mapper.setReporter(value);
		}

		value = unescapeGroup(pattern, "Priority", matcher, repository); //$NON-NLS-1$
		if (value != null && value.length() > 0) {
			for (int n = 0; n < priorities.length; n++) {
				if (priorities[n] != null && priorities[n].contains(value)) {
					mapper.setPriority("P" + (n + 1)); //$NON-NLS-1$
				}
			}
		}
	}

	private Date parseDate(String value) throws ParseException {
		if (date_format != null && date_format.length() > 0) {
			return new SimpleDateFormat(date_format).parse(value);
		}
		return DateFormat.getInstance().parse(value);
	}

	private static String unescapeGroup(NamedPattern pattern, String name,
			Matcher matcher, TaskRepository repository) {
		return unescapeHtml(cleanup(pattern.group(name, matcher), repository));
	}

	// ----------------------------------------------------------------------
	// from org.eclipse.mylyn.internal.web.tasks.WebRepositoryConnector.java
	// ----------------------------------------------------------------------

	private static final String COMPLETED_STATUSES = "completed|fixed|resolved|invalid|verified|deleted|closed|done"; //$NON-NLS-1$
	private final static Date DEFAULT_DATE = new Date(0);

	// rename performQuery to performQueryInternal
	// comment out "static"
	public/* static */IStatus performQueryInternal(String resource,
			String regexp, String taskPrefix, IProgressMonitor monitor,
			TaskDataCollector resultCollector, TaskRepository repository) {
		NamedPattern p = new NamedPattern(regexp, Pattern.CASE_INSENSITIVE
				| Pattern.MULTILINE | Pattern.DOTALL | Pattern.UNICODE_CASE
				| Pattern.CANON_EQ);

		Matcher matcher = p.matcher(resource);

		if (!matcher.find()) {
			return Status.OK_STATUS;
		} else {
			boolean isCorrect = true;
			do {
				if (p.getGroups().isEmpty()) {
					// "classic" mode, no named patterns
					if (matcher.groupCount() < 2) {
						isCorrect = false;
					}
					if (matcher.groupCount() >= 1) {
						String id = matcher.group(1);
						String description = matcher.groupCount() > 1 ? cleanup(
								matcher.group(2), repository)
								: null;
						description = unescapeHtml(description);

						TaskData data = createTaskData(repository, id);
						TaskMapper mapper = new TaskMapper(data, true);
						mapper.setCreationDate(DEFAULT_DATE);
						mapper.setTaskUrl(taskPrefix + id);
						mapper.setSummary(description);
						mapper.setValue(KEY_TASK_PREFIX, taskPrefix);
						resultCollector.accept(data);
					}
				} else {
					String id = p.group("Id", matcher); //$NON-NLS-1$
					String description = p.group("Description", matcher); //$NON-NLS-1$
					if (id == null || description == null) {
						isCorrect = false;
					}
					if (id != null) {
						description = unescapeHtml(description);

						String owner = unescapeHtml(cleanup(p.group("Owner", //$NON-NLS-1$
								matcher), repository));
						String type = unescapeHtml(cleanup(p.group("Type", //$NON-NLS-1$
								matcher), repository));

						TaskData data = createTaskData(repository, id);
						TaskMapper mapper = new TaskMapper(data, true);
						mapper.setCreationDate(DEFAULT_DATE);
						mapper.setTaskUrl(taskPrefix + id);
						mapper.setSummary(description);
						mapper.setValue(KEY_TASK_PREFIX, taskPrefix);
						mapper.setOwner(owner);
						mapper.setTaskKind(type);

						String status = p.group("Status", matcher); //$NON-NLS-1$
						if (status != null) {
							if (COMPLETED_STATUSES.contains(status
									.toLowerCase())) {
								// TODO set actual completion date here
								mapper.setCompletionDate(DEFAULT_DATE);
							}
						}

						// {{ for jp.sourceforge.mylyn.wek.tasks
						updateTaskMapper(mapper, repository, p, matcher);
						// }} for jp.sourceforge.mylyn.wek.tasks

						resultCollector.accept(data);
					}
				}
			} while (matcher.find() && !monitor.isCanceled());

			if (isCorrect) {
				return Status.OK_STATUS;
			} else {
				return new Status(
						IStatus.ERROR,
						TasksWebPlugin.ID_PLUGIN,
						IStatus.ERROR,
						"Require two matching groups (taskId and summary). Check query regexp", //$NON-NLS-1$
						null);
			}
		}
	}

	public IStatus performQuery(TaskRepository repository,
			IRepositoryQuery query, TaskDataCollector resultCollector,
			ISynchronizationSession session, IProgressMonitor monitor) {
		Map<String, String> queryParameters = getQueryParams(query);
		String queryUrl = evaluateParams(query.getUrl(), queryParameters,
				repository);
		try {
			String content = fetchResource(queryUrl, queryParameters,
					repository);

			String taskPrefixAttribute = query.getAttribute(KEY_TASK_PREFIX);
			if (!Util.isPresent(taskPrefixAttribute)) {
				return performRssQuery(content, monitor, resultCollector,
						repository);
			} else {
				String taskPrefix = evaluateParams(taskPrefixAttribute,
						queryParameters, repository);
				String queryPattern = evaluateParams(query
						.getAttribute(KEY_QUERY_PATTERN), queryParameters,
						repository);
				return /* performQuery */performQueryInternal(content,
						queryPattern, taskPrefix, monitor, resultCollector,
						repository);
			}
		} catch (IOException e) {
			String msg = e.getMessage() == null ? e.toString() : e.getMessage();
			return new Status(IStatus.ERROR, TasksWebPlugin.ID_PLUGIN,
					IStatus.ERROR, //
					"Could not fetch resource: " //$NON-NLS-1$
							+ queryUrl + "\n" + msg, e); //$NON-NLS-1$
		}
	}

	private static TaskData createTaskData(TaskRepository taskRepository,
			String id) {
		TaskData data = new TaskData(new TaskAttributeMapper(taskRepository),
				WebRepositoryConnector.REPOSITORY_TYPE, taskRepository
						.getRepositoryUrl(), id);
		data.setPartial(true);
		return data;
	}

	private static String unescapeHtml(String text) {
		if (text == null) {
			return ""; //$NON-NLS-1$
		}
		return StringEscapeUtils.unescapeHtml(text);
	}

	private static String cleanup(String text, TaskRepository repository) {
		if (text == null) {
			return null;
		}

		// Has to disable this for now. See bug 166737 and bug 166936
		// try {
		// text = URLDecoder.decode(text, repository.getCharacterEncoding());
		// } catch (UnsupportedEncodingException ex) {
		// // ignore
		// }

		text = text.replaceAll("<!--.+?-->", ""); //$NON-NLS-1$ //$NON-NLS-2$

		String[] tokens = text.split(" |\\t|\\n|\\r"); //$NON-NLS-1$
		StringBuilder sb = new StringBuilder();
		String sep = ""; //$NON-NLS-1$
		for (String token : tokens) {
			if (token.length() > 0) {
				sb.append(sep).append(token);
				sep = " "; //$NON-NLS-1$
			}
		}

		return sb.toString();
	}
}
